import logging
from pathlib import Path
from typing import Union

from PyQt6.QtCore import pyqtSignal, QFileSystemWatcher
from PyQt6.QtWidgets import QApplication, QVBoxLayout, QWidget
from pyqtgraph.parametertree import Parameter, ParameterTree

from freemocap.parameter_info_models.recording_info_model import (
    RecordingInfoModel,
)
from freemocap.system.paths_and_files_names import get_most_recent_recording_path

logger = logging.getLogger(__name__)


class ActiveRecordingInfoWidget(QWidget):
    new_active_recording_selected_signal = pyqtSignal(RecordingInfoModel)

    def __init__(
        self,
        parent: QWidget = None,
    ):
        super().__init__(parent=parent)
        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._active_recording_info = None
        self._directory_watcher = self._create_directory_watcher()

        self._active_recording_view_widget = ActiveRecordingTreeView(parent=self)
        self._active_recording_view_widget.doubleClicked.connect(self.update_parameter_tree)
        self._layout.addWidget(self._active_recording_view_widget)

    @property
    def active_recording_info(self):
        return self._active_recording_info

    @property
    def active_recording_view_widget(self):
        return self._active_recording_view_widget

    def get_active_recording_info(self, return_path: bool = False):
        # this is redundant to the `active_recording_info` property,
        # but it will be more intuitive to send this down as a callable
        # rather than relying on 'pass-by-reference' magic lol

        if self._active_recording_info is None:
            most_recent_path = get_most_recent_recording_path()
            if most_recent_path is not None:
                self.set_active_recording(most_recent_path)

        if return_path:
            return self._active_recording_info.path

        return self._active_recording_info

    def set_active_recording(
        self,
        recording_folder_path: Union[str, Path],
    ):
        logger.info(f"Setting active recording to {recording_folder_path}")

        self._active_recording_info = RecordingInfoModel(recording_folder_path=str(recording_folder_path))

        self._update_file_watch_path(folder_to_watch=self._active_recording_info.path)

        logger.debug(
            f"Emitting `new_active_recording_selected_signal` for recording: {self._active_recording_info.name}"
        )
        self.new_active_recording_selected_signal.emit(self._active_recording_info)

    def update_parameter_tree(self):
        logger.info("Updating Parameter Tree")
        self._active_recording_view_widget.setup_parameter_tree(self.active_recording_info)

    def _update_file_watch_path(self, folder_to_watch: Union[str, Path]):
        logger.debug(f"Updating file watch path to: {folder_to_watch}")

        if self._directory_watcher.directories():
            logger.debug(f"Removing path from watch list: {self._directory_watcher.directories()}")
            self._directory_watcher.removePaths(self._directory_watcher.directories())

        self._directory_watcher.addPath(folder_to_watch)

    def _create_directory_watcher(self):

        directory_watcher = QFileSystemWatcher()
        directory_watcher.fileChanged.connect(self._handle_directory_changed)
        directory_watcher.directoryChanged.connect(self._handle_directory_changed)
        return directory_watcher

    def _handle_directory_changed(self, path: str):
        logger.info(f"Directory changed: {path} - Updating Parameter Tree")
        self.update_parameter_tree()


class ActiveRecordingTreeView(ParameterTree):
    def __init__(self, parent=None):
        super().__init__(parent=parent, showHeader=False)

    def setup_parameter_tree(
        self,
        recording_info_model: RecordingInfoModel,
    ):
        if recording_info_model is None:
            logger.debug("No recording info model provided - clearing parameter tree")
            self.clear()
            return

        logger.debug(f"Setting up `ActiveRecordingTreeView` for recording: {recording_info_model.name}")
        self.clear()
        self.setParameters(
            self._create_recording_status_parameter_tree(recording_info_model=recording_info_model),
        )

    def _create_recording_status_parameter_tree(self, recording_info_model: RecordingInfoModel):
        parameter_group = Parameter.create(
            name=f"Recording Name: {recording_info_model.name}",
            type="group",
            readonly=True,
            children=[
                dict(
                    name="Path",
                    type="str",
                    value=recording_info_model.path,
                    readonly=True,
                ),
                dict(
                    name="Calibration TOML",
                    type="str",
                    value=recording_info_model.calibration_toml_path,
                    readonly=True,
                ),
                dict(
                    name="Status",
                    type="group",
                    readonly=True,
                    children=[
                        dict(
                            name="Synchronized Videos Recorded",
                            type="str",
                            value=recording_info_model.synchronized_videos_status_check,
                            readonly=True,
                        ),
                        dict(
                            name="2D data exists?",
                            type="str",
                            value=recording_info_model.data2d_status_check,
                            readonly=True,
                        ),
                        dict(
                            name="3D data exists?",
                            type="str",
                            value=recording_info_model.data3d_status_check,
                            readonly=True,
                        ),
                        dict(
                            name="Center-of-Mass data exists?",
                            type="str",
                            value=recording_info_model.center_of_mass_data_status_check,
                            readonly=True,
                        ),
                        dict(
                            name="Blender scene file exists?",
                            type="str",
                            value=recording_info_model.blender_file_status_check,
                            readonly=True,
                        ),
                    ],
                ),
            ],
        )

        return parameter_group


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    active_recording_widget = ActiveRecordingInfoWidget()
    active_recording_widget.show()
    sys.exit(app.exec())



import logging
from pathlib import Path
from typing import Union

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (
    QGroupBox,
    QVBoxLayout,
    QPushButton,
    QRadioButton,
    QWidget,
    QCheckBox,
    QLabel,
    QFileDialog,
    QLineEdit,
)

logger = logging.getLogger(__name__)
BLENDER_EXECUTABLE_PATH_MISSING_STRING = "BLENDER EXECUTABLE NOT FOUND"


class VisualizationControlPanel(QWidget):
    def __init__(self, blender_executable_path: Union[str, Path] = BLENDER_EXECUTABLE_PATH_MISSING_STRING, parent=None):
        super().__init__(parent=parent)

        self._blender_executable_path = str(blender_executable_path)
        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        groupbox = QGroupBox("Export to Blender", parent=self)
        self._layout.addWidget(groupbox)
        vbox = QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.setAlignment(Qt.AlignmentFlag.AlignTop)
        groupbox.setLayout(vbox)

        self._export_to_blender_button = QPushButton("Export to Blender")
        groupbox.layout().addWidget(self._export_to_blender_button)

        self._open_in_blender_automatically_checkbox = QCheckBox("Open in Blender automatically")
        self._open_in_blender_automatically_checkbox.setChecked(True)
        # groupbox.layout().addWidget(self._open_in_blender_automatically_checkbox)

        self._use_default_method_radio_button = QRadioButton("Use default method (Recommended)")
        self._use_default_method_radio_button.setChecked(True)
        groupbox.layout().addWidget(self._use_default_method_radio_button)

        self._use_legacy_method_radio_button = QRadioButton("Use legacy method")
        groupbox.layout().addWidget(self._use_legacy_method_radio_button)

        self._use_cgtinker_method_radio_button = QRadioButton("Use @cgtinker method (Work in progress)")
        groupbox.layout().addWidget(self._use_cgtinker_method_radio_button)

        self._layout.addWidget(QLabel("Blender Executable Path:"))

        self._blender_executable_label = QLineEdit(self._blender_executable_path)
        self._layout.addWidget(self._blender_executable_label)

        self._set_blender_executable_path_button = QPushButton("Choose your Blender Executable")
        self._set_blender_executable_path_button.clicked.connect(self._handle_blender_executable_button_clicked)
        self._layout.addWidget(self._set_blender_executable_path_button)
        self._layout.addStretch()

    @property
    def export_to_blender_button(self):
        return self._export_to_blender_button

    @property
    def open_in_blender_automatically_box_is_checked(self):
        return self._open_in_blender_automatically_checkbox.isChecked()

    @property
    def blender_executable_path(self) -> str:
        """
        Returns the path to the Blender executable

        Returns:
            str: Path to the Blender executable
        """
        return self._blender_executable_path

    def get_user_selected_method_string(self):
        """
        Returns the string that represents the method the user selected to generate the Blender scene.

        Returns:
            str: String representing the method the user selected to generate the Blender scene.
        """
        if self._use_default_method_radio_button.isChecked():
            return "megascript_take2"

        if self._use_legacy_method_radio_button.isChecked():
            return "alpha_megascript"

        if self._use_cgtinker_method_radio_button.isChecked():
            return "cgtinker"

    def _handle_blender_executable_button_clicked(self):
        # from this tutorial - https://www.youtube.com/watch?v=gg5TepTc2Jg&t=649s
        path_selection = QFileDialog.getOpenFileName(
            self,
            "Locate your Blender Executable",
            str(Path().home()),
            "*.*",
        )

        if "blender." in path_selection[0]:
            self._blender_executable_path = path_selection[0]
        elif "Blender.app" in path_selection[0]:
            self._blender_executable_path = path_selection[0] + "/Contents/MacOS/Blender"  # executable is buried on mac
        else:
            self._blender_executable_path = BLENDER_EXECUTABLE_PATH_MISSING_STRING

        logger.info(f"User selected Blender Executable path:{self._blender_executable_path}")
        self._blender_executable_label.setText(self._blender_executable_path)



import logging
from pathlib import Path
from typing import Union

from freemocap.core_processes.export_data.blender_stuff.call_blender_subprocess_methods import (
    call_blender_subprocess_megascript_take2,
    call_blender_subprocess_cgtinker,
    call_blender_subprocess_alpha_megascript,
)
from freemocap.core_processes.export_data.blender_stuff.get_best_guess_of_blender_path import (
    get_best_guess_of_blender_path,
)

logger = logging.getLogger(__name__)


def export_to_blender(
    recording_folder_path: Union[str, Path],
    blender_file_path: Union[str, Path],
    blender_exe_path: Union[str, Path],
    method: str = "megascript_take2",
):

    logger.info(
        f"Exporting session data to a Blender scene at: {str(blender_file_path)} using Blender executable at {str(blender_exe_path)}"
    )

    if method == "megascript_take2":
        call_blender_subprocess_megascript_take2(
            recording_folder_path=recording_folder_path,
            blender_file_path=blender_file_path,
            blender_exe_path=blender_exe_path,
        )
    elif method == "alpha_megascript":
        call_blender_subprocess_alpha_megascript(
            recording_folder_path=recording_folder_path,
            blender_file_path=blender_file_path,
            blender_exe_path=blender_exe_path,
        )
    elif method == "cgtinker":
        call_blender_subprocess_cgtinker(
            recording_folder_path=recording_folder_path,
            blender_file_path=blender_file_path,
            blender_exe_path=blender_exe_path,
        )

    logger.info(f"blender subprocess finished")


if __name__ == "__main__":

    recording_path = r"D:\Dropbox\FreeMoCapProject\FreeMocap_Data\old\sesh_2022-09-19_16_16_50_in_class_jsm"
    export_to_blender(recording_path,
                      recording_path,
                      get_best_guess_of_blender_path())



import logging
from pathlib import Path
from typing import Callable, Union

import numpy as np
from aniposelib.boards import CharucoBoard as AniposeCharucoBoard

from freemocap.core_processes.capture_volume_calibration.anipose_camera_calibration import (
    freemocap_anipose,
)
from freemocap.core_processes.capture_volume_calibration.charuco_stuff.charuco_board_definition import (
    CharucoBoardDefinition,
)
from freemocap.system.paths_and_files_names import (
    get_calibrations_folder_path,
    get_last_successful_calibration_toml_path, create_camera_calibration_file_name,
)

logger = logging.getLogger(__name__)


class AniposeCameraCalibrator:
    def __init__(
        self,
        charuco_board_object: CharucoBoardDefinition,
        charuco_square_size: Union[int, float],
        calibration_videos_folder_path: Union[str, Path],
        progress_callback: Callable[[str], None] = None,
    ):

        self._charuco_board_object = charuco_board_object
        self._progress_callback = progress_callback

        if charuco_square_size == 1:
            logger.warning(
                "Charuco square size is not set, so units of 3d reconstructed data will be in units of `however_long_the_black_edge_of_the_charuco_square_was`. Please input `charuco_square_size` in millimeters (or your preferred unity of length)"
            )
        self._charuco_square_size = charuco_square_size
        self._calibration_videos_folder_path = Path(calibration_videos_folder_path)
        self._recording_folder_path = Path(self._calibration_videos_folder_path).parent
        self._get_video_paths()
        self._initialize_anipose_objects()

    def _get_video_paths(
        self,
    ):
        self._list_of_video_paths = [
            this_video_path for this_video_path in Path(self._calibration_videos_folder_path).glob("*.mp4".lower())
        ]

    def _initialize_anipose_objects(self):
        list_of_camera_names = [this_video_path.stem for this_video_path in self._list_of_video_paths]
        self._anipose_camera_group_object = freemocap_anipose.CameraGroup.from_names(list_of_camera_names)

        # add metadata
        self._anipose_camera_group_object.metadata["charuco_square_size"] = self._charuco_square_size
        self._anipose_camera_group_object.metadata["charuco_board_object"] = str(self._charuco_board_object)
        self._anipose_camera_group_object.metadata["path_to_recorded_videos"] = str(
            self._calibration_videos_folder_path
        )
        self._anipose_camera_group_object.metadata["date_time_calibrated"] = str(np.datetime64("now"))

        self._anipose_charuco_board = AniposeCharucoBoard(
            self._charuco_board_object.number_of_squares_width,
            self._charuco_board_object.number_of_squares_height,
            square_length=self._charuco_square_size,  # mm
            marker_length=self._charuco_square_size * 0.8,
            marker_bits=4,
            dict_size=250,
        )

    def calibrate_camera_capture_volume(self, pin_camera_0_to_origin: bool = False):

        video_paths_list_of_list_of_strings = [[str(this_path)] for this_path in self._list_of_video_paths]

        (error, charuco_frame_data, charuco_frame_numbers,) = self._anipose_camera_group_object.calibrate_videos(
            video_paths_list_of_list_of_strings, self._anipose_charuco_board
        )
        success_str = "Anipose Calibration Successful!"
        logger.info(success_str)
        self._progress_callback(success_str)
        if pin_camera_0_to_origin:
            # translate cameras so camera0 is on `0,0,0`
            self._anipose_camera_group_object = self.pin_camera_zero_to_origin(self._anipose_camera_group_object)
            # self._anipose_camera_group_object = self.rotate_cameras_so_camera_zero_aligns_with_XYZ(self._anipose_camera_group_object)

        # save calibration info to files
        calibration_toml_filename = create_camera_calibration_file_name(recording_name = self._calibration_videos_folder_path.parent.stem)

        calibration_folder_toml_path = Path(get_calibrations_folder_path()) / calibration_toml_filename

        self._anipose_camera_group_object.dump(calibration_folder_toml_path)
        logger.info(f"anipose camera calibration data saved to Calibrations folder - {str(calibration_folder_toml_path)}")

        recording_folder_toml_path = Path(self._recording_folder_path) / calibration_toml_filename

        self._anipose_camera_group_object.dump(recording_folder_toml_path)
        logger.info(f"anipose camera calibration data saved to Recording folder - {str(recording_folder_toml_path)}")

        last_successful_calibration_toml_path = get_last_successful_calibration_toml_path()
        self._anipose_camera_group_object.dump(last_successful_calibration_toml_path)

        logger.info(f"anipose camera calibration data also saved to 'Last Successful Calibration' - {last_successful_calibration_toml_path}")


    def pin_camera_zero_to_origin(self, _anipose_camera_group_object):
        original_translation_vectors = _anipose_camera_group_object.get_translations()
        camera_0_translation = original_translation_vectors[0, :]
        altered_translation_vectors = np.zeros(original_translation_vectors.shape)
        for this_camera_number in range(original_translation_vectors.shape[0]):
            altered_translation_vectors[this_camera_number, :] = (
                original_translation_vectors[this_camera_number, :] - camera_0_translation
            )

        _anipose_camera_group_object.set_translations(altered_translation_vectors)
        logger.info(f"original translation vectors:\n {original_translation_vectors}")
        logger.info(f"altered translation vectors:\n {_anipose_camera_group_object.get_translations()}")
        return _anipose_camera_group_object

    def rotate_cameras_so_camera_zero_aligns_with_XYZ(self, _anipose_camera_group_object):
        logger.warning("this function does not work, dont use it lol ")
        pass
        # original_rotations_euler = _anipose_camera_group_object.get_rotations()
        # original_translation_vectors = _anipose_camera_group_object.get_translations()
        # camera_rotation_matrix_list = [
        #     Rotation.from_euler('xyz', original_rotations_euler[this_cam_num, :]).as_matrix() for
        #     this_cam_num in range(original_rotations_euler.shape[0])]
        #
        # rotated_translation_vectors = [camera_rotation_matrix_list[0] @ this_tx for this_tx in
        #                                original_translation_vectors]
        # _anipose_camera_group_object.set_rotations(rotated_translation_vectors)
        # return _anipose_camera_group_object



from PyQt6.QtGui import QAction, QDesktopServices
from PyQt6.QtCore import QUrl


CREATE_NEW_RECORDING_ACTION_NAME = "New Recording"
LOAD_MOST_RECENT_RECORDING_ACTION_NAME = "Load Most Recent Recording"
LOAD_RECORDING_ACTION_NAME = "Load Recording"
IMPORT_VIDEOS_ACTION_NAME = "Import Videos"
KILL_THREADS_AND_PROCESSES_ACTION_NAME = "Kill Threads and Processes"
REBOOT_GUI_ACTION_NAME = "Reboot GUI"
EXIT_ACTION_NAME = "Exit"

OPEN_DOCS_ACTION_NAME = "Open Documentation"
FREEMOCAP_FOUNDATION_ACTION_NAME = "Freemocap Foundation"

DONATE_ACTION_NAME = "Donate to Freemocap"


class Actions:
    def __init__(self, freemocap_main_window):
        # File
        self.create_new_recording_action = QAction(CREATE_NEW_RECORDING_ACTION_NAME, parent=freemocap_main_window)
        self.create_new_recording_action.setShortcut("Ctrl+N")
        self.create_new_recording_action.triggered.connect(freemocap_main_window.handle_start_new_session_action)

        self.load_most_recent_recording_action = QAction(
            LOAD_MOST_RECENT_RECORDING_ACTION_NAME, parent=freemocap_main_window
        )
        self.load_most_recent_recording_action.setShortcut("Ctrl+D")
        self.load_most_recent_recording_action.triggered.connect(
            freemocap_main_window.handle_load_most_recent_recording
        )

        self.load_existing_recording_action = QAction(LOAD_RECORDING_ACTION_NAME, parent=freemocap_main_window)
        self.load_existing_recording_action.setShortcut("Ctrl+O")
        self.load_existing_recording_action.triggered.connect(freemocap_main_window.open_load_existing_recording_dialog)

        self.import_videos_action = QAction(IMPORT_VIDEOS_ACTION_NAME, parent=freemocap_main_window)
        self.import_videos_action.setShortcut("Ctrl+I")
        self.import_videos_action.triggered.connect(freemocap_main_window.open_import_videos_dialog)

        self.reboot_gui_action = QAction(REBOOT_GUI_ACTION_NAME, parent=freemocap_main_window)
        self.reboot_gui_action.setShortcut("Ctrl+R")
        self.reboot_gui_action.triggered.connect(freemocap_main_window.reboot_gui)

        self.kill_running_threads_and_processes_action = QAction(
            KILL_THREADS_AND_PROCESSES_ACTION_NAME, parent=freemocap_main_window
        )
        self.kill_running_threads_and_processes_action.setShortcut("Ctrl+K")
        self.kill_running_threads_and_processes_action.triggered.connect(
            freemocap_main_window.kill_running_threads_and_processes
        )

        self.exit_action = QAction("E&xit", parent=freemocap_main_window)
        self.exit_action.setShortcut("Ctrl+Q")
        self.exit_action.triggered.connect(freemocap_main_window.close)

        # Help
        self.open_docs_action = QAction(OPEN_DOCS_ACTION_NAME, parent=freemocap_main_window)
        self.open_docs_action.triggered.connect(lambda: QDesktopServices.openUrl(QUrl("https://freemocap.readthedocs.io/en/latest/")))
        
        self.freemocap_foundation_action = QAction(FREEMOCAP_FOUNDATION_ACTION_NAME, parent=freemocap_main_window)
        self.freemocap_foundation_action.triggered.connect(lambda: QDesktopServices.openUrl(QUrl("https://freemocap.org/about-us.html")))

        # # Navigation
        # show_camera_control_panel_action = QAction("&1 - Show Camera Control Panel", parent=main_window)
        # show_camera_control_panel_action.setShortcut("Ctrl+1")
        #
        # show_calibrate_capture_volume_panel_action = QAction(
        #     "&2 - Show Calibrate Capture Volume Panel", parent=main_window
        # )
        # show_calibrate_capture_volume_panel_action.setShortcut("Ctrl+2")
        #
        # show_motion_capture_videos_panel_action = QAction("&3 - Show Motion Capture Videos Panel", parent=main_window)
        # show_motion_capture_videos_panel_action.setShortcut("Ctrl+3")
        #
        # Support
        self.donate_action = QAction(DONATE_ACTION_NAME, parent=freemocap_main_window)
        self.donate_action.triggered.connect(lambda: QDesktopServices.openUrl(QUrl("https://freemocap.org/about-us.html#donate")))
        # self.send_usage_statistics_action = QAction("Send &User Statistics", parent=freemocap_main_window)
        # self.user_survey_action = QAction("&User Survey", parent=freemocap_main_window)



from mediapipe.python.solutions import holistic as mp_holistic

mediapipe_body_connections = [connection for connection in mp_holistic.POSE_CONNECTIONS]

mediapipe_hand_connections = [connection for connection in mp_holistic.HAND_CONNECTIONS]

mediapipe_face_connections = [connection for connection in mp_holistic.FACEMESH_CONTOURS]

mediapipe_body_landmark_names = [landmark.name.lower() for landmark in mp_holistic.PoseLandmark]
NUMBER_OF_MEDIAPIPE_BODY_MARKERS = len(mediapipe_body_landmark_names)

mediapipe_hand_landmark_names = [landmark.name.lower() for landmark in mp_holistic.HandLandmark]


mediapipe_face_landmark_names = [
    "right_eye",
    "left_eye",
    "nose_tip",
    "mouth_center",
    "right_ear_tragion",
    "left_ear_tragion",
]

mediapipe_tracked_point_names_dict = {
    "body": mediapipe_body_landmark_names,
    "right_hand": [f"right_hand_{name}" for name in mediapipe_hand_landmark_names],
    "left_hand": [f"left_hand_{name}" for name in mediapipe_hand_landmark_names],
    "face": mediapipe_face_landmark_names,
}

mediapipe_names_and_connections_dict = {
    "body": {
        "names": mediapipe_body_landmark_names,
        "connections": mediapipe_body_connections,
    },
    "right_hand": {
        "names": [f"right_hand_{name}" for name in mediapipe_hand_landmark_names],
        "connections": mediapipe_hand_connections,
    },
    "left_hand": {
        "names": [f"left_hand_{name}" for name in mediapipe_hand_landmark_names],
        "connections": mediapipe_hand_connections,
    },
    "face": {
        "names": mediapipe_face_landmark_names,
        "connections": mediapipe_face_connections,
    },
}


if __name__ == "__main__":
    import pprint

    pprint.pp(mediapipe_tracked_point_names_dict)



import logging

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QPixmap
from PyQt6.QtWidgets import (
    QCheckBox,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from freemocap.gui.qt.actions_and_menus.actions import (
    CREATE_NEW_RECORDING_ACTION_NAME,
    LOAD_MOST_RECENT_RECORDING_ACTION_NAME,
    LOAD_RECORDING_ACTION_NAME,
    IMPORT_VIDEOS_ACTION_NAME,
    Actions,
)
from freemocap.system.paths_and_files_names import PATH_TO_FREEMOCAP_LOGO_SVG, SPARKLES_EMOJI_STRING

logger = logging.getLogger(__name__)


class WelcomeScreenButton(QPushButton):
    def __init__(self, text: str, parent: QWidget = None):
        super().__init__(text, parent=parent)
        self.setFixedHeight(50)
        self.setFixedWidth(400)


class HomeWidget(QWidget):
    def __init__(self, actions: Actions, parent: QWidget = None):
        super().__init__(parent=parent)

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self.sizePolicy().setHorizontalStretch(1)
        self.sizePolicy().setVerticalStretch(1)

        self._layout.addStretch(1)

        self._add_freemocap_logo()

        self._welcome_to_freemocap_title_widget = self._welcome_to_freemocap_title()
        self._layout.addWidget(self._welcome_to_freemocap_title_widget)

        self._create_new_session_button = WelcomeScreenButton(
            f"{CREATE_NEW_RECORDING_ACTION_NAME}"
        )
        self._create_new_session_button.clicked.connect(actions.create_new_recording_action.trigger)
        self._layout.addWidget(self._create_new_session_button, alignment=Qt.AlignmentFlag.AlignCenter)
        self._create_new_session_button.setProperty("recommended_next", True)

        self._load_most_recent_session_button = WelcomeScreenButton(
            f"{LOAD_MOST_RECENT_RECORDING_ACTION_NAME}"
        )
        # self._load_most_recent_session_button.clicked.connect(actions.load_most_recent_recording_action.trigger)
        # self._layout.addWidget(self._load_most_recent_session_button, alignment=Qt.AlignmentFlag.AlignCenter)
        #
        self._load_existing_session_button = WelcomeScreenButton(f"{LOAD_RECORDING_ACTION_NAME}")
        self._load_existing_session_button.clicked.connect(actions.load_existing_recording_action.trigger)
        self._layout.addWidget(self._load_existing_session_button, alignment=Qt.AlignmentFlag.AlignCenter)

        self._import_videos_button = WelcomeScreenButton(f"{IMPORT_VIDEOS_ACTION_NAME}")
        self._import_videos_button.clicked.connect(actions.import_videos_action.trigger)
        self._layout.addWidget(self._import_videos_button, alignment=Qt.AlignmentFlag.AlignCenter)

        self._layout.addStretch(1)

        send_pings_string = "Send anonymous usage information"
        self._send_pings_checkbox = QCheckBox(send_pings_string)
        self._send_pings_checkbox.setChecked(True)
        self._layout.addWidget(self._send_pings_checkbox, alignment=Qt.AlignmentFlag.AlignCenter)

        privacy_policy_link_string = '<a href="https://freemocap.readthedocs.io/en/latest/privacy_policy/" style="color: #333333;">Click here to view our privacy policy</a>'
        self._privacy_policy_link = QLabel(privacy_policy_link_string)
        self._privacy_policy_link.setOpenExternalLinks(True)
        self._layout.addWidget(self._privacy_policy_link, alignment=Qt.AlignmentFlag.AlignCenter)

        self.style().polish(self)

    def _welcome_to_freemocap_title(self):
        # TO DO - this shouldn't be part of the `camera_view_panel` - it should be its own thing that gets swapped out on session start
        logger.info("Creating `welcome to freemocap` layout")

        session_title_label = QLabel("Welcome  to  FreeMoCap!")
        session_title_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        session_title_label.setStyleSheet("font-size: 54px;")

        return session_title_label

    def _add_freemocap_logo(self):
        freemocap_logo_label = QLabel(self)
        freemocap_logo_label.sizePolicy().setHorizontalStretch(1)
        freemocap_logo_label.sizePolicy().setVerticalStretch(1)
        self._layout.addWidget(freemocap_logo_label)
        freemocap_logo_pixmap = QPixmap(PATH_TO_FREEMOCAP_LOGO_SVG)
        freemocap_logo_pixmap = freemocap_logo_pixmap.scaledToWidth(200)
        freemocap_logo_label.setPixmap(freemocap_logo_pixmap)
        freemocap_logo_label.setAlignment(Qt.AlignmentFlag.AlignCenter)



from pathlib import Path
from typing import Union

import numpy as np

from freemocap.tests.utilities.get_number_of_frames_of_videos_in_a_folder import (
    get_number_of_frames_of_videos_in_a_folder,
)


def test_total_body_center_of_mass_data_shape(
    synchronized_videos_folder: Union[str, Path],
    total_body_center_of_mass_npy_file_path: Union[str, Path],
):

    assert Path(
        total_body_center_of_mass_npy_file_path
    ).is_file(), f"No file found at {total_body_center_of_mass_npy_file_path}"

    total_body_center_of_mass_fr_xyz = np.load(total_body_center_of_mass_npy_file_path)

    frame_count = get_number_of_frames_of_videos_in_a_folder(synchronized_videos_folder)
    assert (
        len(set(frame_count)) == 1
    ), f"Videos in {synchronized_videos_folder} have different frame counts: {frame_count}"

    number_of_frames = frame_count[0]

    assert (
        total_body_center_of_mass_fr_xyz.shape[0] == number_of_frames
    ), f"Number of frames in {total_body_center_of_mass_npy_file_path} does not match number of frames of videos in {synchronized_videos_folder}"

    assert (
        total_body_center_of_mass_fr_xyz.shape[1] == 3
    ), f"`total_body_center_of_mass_fr_xyz.shape[1]` should have 3 dimensions (X,Y,Z) "

    return True



from pyqtgraph.parametertree import Parameter, ParameterTree

from freemocap.parameter_info_models.recording_processing_parameter_models import MediapipeParametersModel, \
    RecordingProcessingParameterModel, AniposeTriangulate3DParametersModel, PostProcessingParametersModel, \
    ButterworthFilterParametersModel

BUTTERWORTH_ORDER = "Order"

BUTTERWORTH_CUTOFF_FREQUENCY = "Cutoff Frequency"

POST_PROCESSING_FRAME_RATE = "framerate"

BUTTERWORTH_FILTER_TREE_NAME = "Butterworth Filter"

USE_RANSAC_METHOD = "Use RANSAC Method"

ANIPOSE_CONFIDENCE_CUTOFF = "Confidence Threshold Cut-off"

ANIPOSE_TREE_NAME = "Anipose Triangulation"

STATIC_IMAGE_MODE = "Static Image Mode"

MINIUMUM_TRACKING_CONFIDENCE = "Minimum Tracking Confidence"

MINIMUM_DETECTION_CONFIDENCE = "Minimum Detection Confidence"

MODEL_COMPLEXITY = "Model Complexity"

MEDIAPIPE_TREE_NAME = "Mediapipe"


def create_mediapipe_parameter_group(
        parameter_model: MediapipeParametersModel = MediapipeParametersModel(),
) -> Parameter:
    model_complexity_list = [
        "0 (Fastest/Least accurate)",
        "1 (Middle ground)",
        "2 (Slowest/Most accurate)",
    ]
    return Parameter.create(
        name=MEDIAPIPE_TREE_NAME,
        type="group",
        children=[
            dict(
                name=MODEL_COMPLEXITY,
                type="list",
                limits=model_complexity_list,
                value=model_complexity_list[parameter_model.model_complexity],
                tip="Which Mediapipe model to use - higher complexity is slower but more accurate. "
                    "Variable name in `mediapipe` code: `model_complexity`",
            ),
            dict(
                name=MINIMUM_DETECTION_CONFIDENCE,
                type="float",
                value=parameter_model.min_detection_confidence,
                step=0.05,
                limits=(0.0, 1.0),
                tip="Minimum confidence for a skeleton detection to be considered valid. "
                    "Variable name in `mediapipe` code: `min_detection_confidence`."
                    "NOTE - Never trust a machine learning model's estimates of their own confidence!",
            ),
            dict(
                name=MINIUMUM_TRACKING_CONFIDENCE,
                type="float",
                value=parameter_model.min_tracking_confidence,
                step=0.05,
                limits=(0.0, 1.0),
                tip="Minimum confidence needed to use the previous frame's skeleton estiamte to predict the next one"
                    "Variable name in `mediapipe` code: `min_tracking_confidence`.",
            ),
            dict(
                name=STATIC_IMAGE_MODE,
                type="bool",
                value=parameter_model.static_image_mode,
                tip="If true, the model will process each image independently, without tracking across frames."
                    "I think this is equivalent to setting `min_tracking_confidence` to 0.0"
                    "Variable name in `mediapipe` code: `static_image_mode`",
            ),
        ],
    )


def create_3d_triangulation_prarameter_group(
        parameter_model: AniposeTriangulate3DParametersModel = AniposeTriangulate3DParametersModel(),
) -> Parameter:
    return Parameter.create(
        name=ANIPOSE_TREE_NAME,
        type="group",
        children=[
            dict(
                name=ANIPOSE_CONFIDENCE_CUTOFF,
                type="float",
                value=parameter_model.confidence_threshold_cutoff,
                tip="Confidence threshold cut-off for triangulation. "
                    "NOTE - Never trust a machine learning model's estimates of their own confidence! "
                    "TODO - Something similar that uses `reprojection_error` instead of `confidence`",
            ),
            dict(
                name=USE_RANSAC_METHOD,
                type="bool",
                value=parameter_model.use_triangulate_ransac_method,
                tip="If true, use `anipose`'s `triangulate_ransac` method instead of the default `triangulate_simple` method. "
                    "NOTE - Much slower than the 'simple' method, but might be more accurate and better at rejecting bad camera views. Needs more testing and evaluation to see if it's worth it. ",
            ),
        ],
    )


def create_post_processing_parameter_group(
        parameter_model: PostProcessingParametersModel = PostProcessingParametersModel(),
) -> Parameter:
    return Parameter.create(
        name=BUTTERWORTH_FILTER_TREE_NAME,
        type="group",
        children=[
            dict(
                name=POST_PROCESSING_FRAME_RATE,
                type="float",
                value=parameter_model.butterworth_filter_parameters.sampling_rate,
                tip="Framerate of the recording " "TODO - Calculate this from the recorded timestamps....",
            ),
            dict(
                name=BUTTERWORTH_CUTOFF_FREQUENCY,
                type="float",
                value=parameter_model.butterworth_filter_parameters.cutoff_frequency,
                tip="Oscillations above this frequency will be filtered from the data. ",
            ),
            dict(
                name=BUTTERWORTH_ORDER,
                type="int",
                value=parameter_model.butterworth_filter_parameters.order,
                tip="Order of the filter."
                    "NOTE - I'm not really sure what this parameter does, but this is what I see in other people's Methods sections so....   lol",
            ),
        ],
        tip="Low-pass, zero-lag, Butterworth filter to remove high frequency oscillations/noise from the data. ",
    )


def extract_mediapipe_parameter_model_from_parameter_tree(
        parameter_tree: ParameterTree,
) -> RecordingProcessingParameterModel:
    return RecordingProcessingParameterModel(
        mediapipe_parameters_model=MediapipeParametersModel(
            model_complexity=parameter_tree.param(MEDIAPIPE_TREE_NAME).param(MODEL_COMPLEXITY).value(),
            min_detection_confidence=parameter_tree.param(MEDIAPIPE_TREE_NAME).param(
                MINIMUM_DETECTION_CONFIDENCE).value(),
            min_tracking_confidence=parameter_tree.param(MEDIAPIPE_TREE_NAME).param(
                MINIUMUM_TRACKING_CONFIDENCE).value(),
            static_image_mode=parameter_tree.param(MEDIAPIPE_TREE_NAME).param(STATIC_IMAGE_MODE).value(), ),

        anipose_triangulate_3d_parameters_model=AniposeTriangulate3DParametersModel(
            confidence_threshold_cutoff=parameter_tree.param(ANIPOSE_TREE_NAME).param(
                ANIPOSE_CONFIDENCE_CUTOFF).value(),
            use_triangulate_ransac_method=parameter_tree.param(ANIPOSE_TREE_NAME).param(USE_RANSAC_METHOD).value(),
            ),
        post_processing_parameters_model=PostProcessingParametersModel(
            framerate=parameter_tree.param(BUTTERWORTH_FILTER_TREE_NAME).param(POST_PROCESSING_FRAME_RATE).value(),
            butterworth_filter_parameters=ButterworthFilterParametersModel(
                sampling_rate=parameter_tree.param(BUTTERWORTH_FILTER_TREE_NAME).param(
                    POST_PROCESSING_FRAME_RATE).value(),
                cutoff_frequency=parameter_tree.param(BUTTERWORTH_FILTER_TREE_NAME).param(
                    BUTTERWORTH_CUTOFF_FREQUENCY).value(),
                order=parameter_tree.param(BUTTERWORTH_FILTER_TREE_NAME).param(BUTTERWORTH_ORDER).value(),
                ),
            )
        ) #thanks github co-pilot lol



import logging

from pydantic import BaseModel

from freemocap.parameter_info_models.recording_info_model import (
    RecordingInfoModel,
)

logger = logging.getLogger(__name__)


class MediapipeParametersModel(BaseModel):
    model_complexity: int = 2
    min_detection_confidence: float = 0.5
    min_tracking_confidence: float = 0.5
    static_image_mode: bool = False


class AniposeTriangulate3DParametersModel(BaseModel):
    confidence_threshold_cutoff: float = 0.7
    use_triangulate_ransac_method: bool = False


class ButterworthFilterParametersModel(BaseModel):
    sampling_rate: float = 30
    cutoff_frequency: float = 7
    order: int = 4


class PostProcessingParametersModel(BaseModel):
    framerate: float = 30.0
    butterworth_filter_parameters = ButterworthFilterParametersModel()
    max_gap_to_fill: int = 10


class RecordingProcessingParameterModel(BaseModel):
    recording_info_model: RecordingInfoModel = None
    mediapipe_parameters_model: MediapipeParametersModel = MediapipeParametersModel()
    anipose_triangulate_3d_parameters_model: AniposeTriangulate3DParametersModel = AniposeTriangulate3DParametersModel()
    post_processing_parameters_model: PostProcessingParametersModel = PostProcessingParametersModel()

    class Config:
        arbitrary_types_allowed = True



import logging
import logging.handlers
import sys
from logging.config import dictConfig

from freemocap.system.paths_and_files_names import get_log_file_path

DEFAULT_LOGGING = {"version": 1, "disable_existing_loggers": False}

LOG_FILE_PATH = None

format_string = "[%(asctime)s.%(msecs)04d] [%(levelname)8s] [%(name)s] [%(funcName)s():%(lineno)s] [PID:%(process)d TID:%(thread)d] %(message)s"

default_logging_formatter = logging.Formatter(fmt=format_string, datefmt="%Y-%m-%d %H:%M:%S")


def get_logging_handlers():
    dictConfig(DEFAULT_LOGGING)

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.DEBUG)
    console_handler.setFormatter(default_logging_formatter)

    # Setup File handler (from https://stackoverflow.com/a/24507130/14662833 )

    file_handler = logging.FileHandler(get_log_file_path())
    file_handler.setFormatter(default_logging_formatter)
    file_handler.setLevel(logging.DEBUG)

    return [console_handler, file_handler]


def configure_logging():
    print(f"Setting up freemocap logging  {__file__}")

    if len(logging.getLogger().handlers) == 0:
        handlers = get_logging_handlers()
        for handler in handlers:
            if not handler in logging.getLogger("").handlers:
                logging.getLogger("").handlers.append(handler)

        logging.root.setLevel(logging.DEBUG)
    else:
        logger = logging.getLogger(__name__)
        logger.info("Logging already configured")



import json
import logging
import sys
from copy import copy
from pathlib import Path
from typing import List, Union, Dict

import addon_utils
import bpy
import numpy as np


print("Running script to create Blender file from freemocap session data: " + __file__)

logger = logging.getLogger(__name__)


###############################################################
### IF YOU ARE RUNNING THIS SCRIPT FROM WITHIN BLENDER -
### CHANGE THE PATH TO THE SESSION FOLDER IN THE `if __name__ == "__main__":` BLOCK AT THE **BOTTOM** OF THIS FILE
### the variable is named `recording_path_input`
###############################################################


def main(recording_path: Union[str, Path],
         blender_file_save_path: Union[str, Path],
         mediapipe_empty_names: Dict[str, List[str]],
         create_rig: bool = True,
         ):
    ###%% clear the scene - Scorch the earth \o/
    print("Clearing scene...")
    try:
        bpy.ops.object.mode_set(mode="OBJECT")
    except:
        pass
    try:
        bpy.ops.object.select_all(action="SELECT")  # select all objects
        bpy.ops.object.delete(use_global=True)  # delete all objects from all scenes
    except:
        pass

    print("Loading data....")

    # paths and whatnot
    try:
        # load mediapipe skeleton data
        path_to_data_arrays_folder = recording_path / "output_data"
        path_to_body_npy = path_to_data_arrays_folder / "mediapipe_body_3d_xyz.npy"
        path_to_right_hand_npy = path_to_data_arrays_folder / "mediapipe_right_hand_3d_xyz.npy"
        path_to_left_hand_npy = path_to_data_arrays_folder / "mediapipe_left_hand_3d_xyz.npy"
        path_to_face_npy = path_to_data_arrays_folder / "mediapipe_face_3d_xyz.npy"

        mediapipe_body_fr_mar_xyz = np.load(str(path_to_body_npy)) / 1000  # load and convert to meters
        mediapipe_right_hand_fr_mar_xyz = np.load(str(path_to_right_hand_npy)) / 1000
        mediapipe_left_hand_fr_mar_xyz = np.load(str(path_to_left_hand_npy)) / 1000
        mediapipe_face_fr_mar_xyz = np.load(str(path_to_face_npy)) / 1000

        # load reprojection error
        path_to_reprojection_error_npy = (
                path_to_data_arrays_folder / "raw_data" / "mediapipe3dData_numFrames_numTrackedPoints_reprojectionError.npy"
        )
        mediapipe_reprojection_error_fr_mar = np.load(str(path_to_reprojection_error_npy))

        # load skeleton segment lengths
        path_to_segment_length_json = path_to_data_arrays_folder / "mediapipe_skeleton_segment_lengths.json"
        print(f"loading skeleton segment lengths from `json` at - {str(path_to_segment_length_json)}")
        f = open(path_to_segment_length_json)
        skeleton_segment_lengths_dict = json.load(f)

    except Exception as e:
        print("Failed to load freemocap data")
        raise e

    number_of_frames = mediapipe_body_fr_mar_xyz.shape[0]
    print(f"mediapipe_body_fr_mar_dim.shape: {mediapipe_body_fr_mar_xyz.shape}")
    print(f"mediapipe_reprojection_error_fr_mar.shape: {mediapipe_reprojection_error_fr_mar.shape}")

    #########################
    ### Create Origin Axes

    bpy.ops.object.empty_add(type="SPHERE", scale=(3.0, 3.0, 3.0))
    world_origin_axes = bpy.context.editable_objects[-1]
    world_origin_axes.name = "world_origin"  # will stay at origin

    bpy.ops.object.empty_add(type="ARROWS")
    freemocap_origin_axes = bpy.context.editable_objects[-1]
    freemocap_origin_axes.name = (
        "freemocap_origin_axes"  # will translate to put skelly on ground symmetric-ish about origin
    )

    ##############################
    # %% Set start and end frames

    start_frame = 1
    end_frame = number_of_frames

    bpy.context.scene.frame_start = start_frame
    bpy.context.scene.frame_end = end_frame

    ############################
    ### Load mocap data as empty markers

    print("_________________________\n"
          "__________________________\n"
          "Loading freemocap trajectory data as empty markers..."
          "_________________________\n")

    mediapipe_body_trajectory_names = [empty_name for empty_name in mediapipe_empty_names["body"]]

    mediapipe_right_hand_trajectory_names = [f"right_hand_{empty_name}" for empty_name in mediapipe_empty_names["hand"]]
    mediapipe_left_hand_trajectory_names = [f"left_hand_{empty_name}" for empty_name in mediapipe_empty_names["hand"]]
    # mediapipe_face_trajectory_names = [f"face_{number}:{empty_name}" for number, empty_name in
    #                                    face_contour_marker_indices]

    # create empty markers for body
    body_empty_scale = 0.02
    for marker_number in range(mediapipe_body_fr_mar_xyz.shape[1]):
        trajectory_name = mediapipe_body_trajectory_names[marker_number]
        trajectory_fr_xyz = mediapipe_body_fr_mar_xyz[:, marker_number, :]
        create_keyframed_empty_from_3d_trajectory_data(
            trajectory_fr_xyz=trajectory_fr_xyz,
            trajectory_name=trajectory_name,
            parent_origin=freemocap_origin_axes,
            empty_scale=0.01,
            empty_type="SPHERE",
        )

    # create empty markers for right hand
    hand_empty_scale = 0.01
    for marker_number in range(mediapipe_right_hand_fr_mar_xyz.shape[1]):
        trajectory_name = mediapipe_right_hand_trajectory_names[marker_number]
        trajectory_fr_xyz = mediapipe_right_hand_fr_mar_xyz[:, marker_number, :]
        create_keyframed_empty_from_3d_trajectory_data(
            trajectory_fr_xyz=trajectory_fr_xyz,
            trajectory_name=trajectory_name,
            parent_origin=freemocap_origin_axes,
            empty_scale=hand_empty_scale,
            empty_type="PLAIN_AXES",
        )

    # create empty markers for left hand
    for marker_number in range(mediapipe_left_hand_fr_mar_xyz.shape[1]):
        trajectory_name = mediapipe_left_hand_trajectory_names[marker_number]
        trajectory_fr_xyz = mediapipe_left_hand_fr_mar_xyz[:, marker_number, :]
        create_keyframed_empty_from_3d_trajectory_data(
            trajectory_fr_xyz=trajectory_fr_xyz,
            trajectory_name=trajectory_name,
            parent_origin=freemocap_origin_axes,
            empty_scale=hand_empty_scale,
            empty_type="PLAIN_AXES",
        )

    # # create empty markers for face
    # face_empty_scale = 0.005
    # for marker_number in face_contour_marker_indices:
    #
    #     if marker_number < len(mediapipe_face_trajectory_names):
    #         trajectory_name = mediapipe_face_trajectory_names[marker_number]
    #     else:
    #         trajectory_name = f"face_{marker_number}"
    #
    #     trajectory_fr_xyz = mediapipe_face_fr_mar_xyz[:, marker_number, :]
    #     create_keyframed_empty_from_3d_trajectory_data(
    #         trajectory_fr_xyz=trajectory_fr_xyz,
    #         trajectory_name=trajectory_name,
    #         parent_origin=freemocap_origin_axes,
    #         empty_scale=face_empty_scale,
    #         empty_type="PLAIN_AXES",
    #     )

    #######################################################################
    # %% create virtual markers
    print("_________________________\n"
          "-------------------------\n"
          "Creating virtual markers...")

    # verify that the virtual marker definition dictionary is valid
    test_virtual_marker_definitions(mediapipe_virtual_marker_definitions_dict)

    virtual_marker_names = list(mediapipe_virtual_marker_definitions_dict.keys())
    for virtual_marker_name in virtual_marker_names:
        component_trajectory_names = mediapipe_virtual_marker_definitions_dict[virtual_marker_name]["marker_names"]
        trajectory_weights = mediapipe_virtual_marker_definitions_dict[virtual_marker_name]["marker_weights"]

        print(
            f"Calculating virtual marker trajectory: {virtual_marker_name} \n"
            f"Component trajectories: {component_trajectory_names} \n"
            f" weights: {trajectory_weights}\n")

        virtual_marker_xyz = calculate_virtual_marker_trajectory(
            trajectory_3d_frame_marker_xyz=mediapipe_body_fr_mar_xyz,
            all_trajectory_names=mediapipe_body_trajectory_names,
            component_trajectory_names=component_trajectory_names,
            trajectory_weights=trajectory_weights,
        )
        create_keyframed_empty_from_3d_trajectory_data(
            trajectory_fr_xyz=virtual_marker_xyz,
            trajectory_name=virtual_marker_name,
            parent_origin=freemocap_origin_axes,
            empty_scale=body_empty_scale * 3,
            empty_type="PLAIN_AXES",
        )

    print(f"Adding virtual marker names to body trajectory names  - {virtual_marker_names}")
    mediapipe_body_trajectory_names.extend(virtual_marker_names)
    print("Done creating virtual markers")

    #######################################################################
    # put sphere meshes on empty markers for visualization

    # body
    put_sphere_meshes_on_empties(
        empty_names_list=mediapipe_body_trajectory_names,
        parent_object=freemocap_origin_axes,
        sphere_scale=body_empty_scale,
    )

    # create_stick_meshes_between_emtpies(
    #     body_frame_marker_xyz=mediapipe_body_fr_mar_xyz,
    #     empty_names=mediapipe_body_trajectory_names,
    #     body_connections=body_connections,
    #     parent_object=freemocap_origin_axes,
    # )

    # right hand
    put_sphere_meshes_on_empties(
        empty_names_list=mediapipe_right_hand_trajectory_names,
        parent_object=freemocap_origin_axes,
        sphere_scale=hand_empty_scale,
    )

    put_sphere_meshes_on_empties(
        empty_names_list=mediapipe_left_hand_trajectory_names,
        parent_object=freemocap_origin_axes,
        sphere_scale=hand_empty_scale,
    )

    # # face
    # put_sphere_meshes_on_empties(
    #     empty_names_list=mediapipe_face_trajectory_names,
    #     parent_object=freemocap_origin_axes,
    #     sphere_scale=face_empty_scale,
    # )

    ###########################
    ### Make stick figure
    # creating_stick_figure_mesh_from_bone_dictionary(trajectory_frame_marker_xyz=mediapipe_body_fr_mar_xyz,
    #                                                 body_connections=body_connections,)
    ############################
    ### Load videos into scene
    if Path(recording_path / "annotated_videos").is_dir():
        videos_path = Path(recording_path / "annotated_videos")
    elif Path(recording_path / "synchronized_videos").is_dir():
        videos_path = Path(recording_path / "synchronized_videos")
    else:
        print("Did not find an `annotated_videos` or `synchronized_videos` folder in the recording path")
        videos_path = None

    if videos_path is not None:
        try:
            addon_utils.enable("io_import_images_as_planes")
        except Exception as e:
            print("Error enabling `io_import_images_as_planes` addon: ")
            print(e)
        try:
            add_videos_to_scene(videos_path=videos_path, parent_object=world_origin_axes)
        except Exception as e:
            print("Error adding videos to scene: ")
            print(e)

    if create_rig:
        try:
            #######################################################################
            ### Create meta-rig
            print("Saving blender file with raw motion capture data before moving on to rigging")
            bpy.ops.object.mode_set(mode="OBJECT")
            bpy.ops.wm.save_as_mainfile(filepath=str(blender_file_save_path))

            print(
                "____________________________________________________________________________\n",
                "Done loading in motion capture data -  Now lets use it to drive an armature!\n"
                "____________________________________________________________________________\n",
            )

            #######################################################################
            ##% Activate necessary addons
            addon_utils.enable("rigify")

            ######################
            ### Create Rigify human meta-rig
            print(f"Creating `rigify human meta-rig`")
            bpy.ops.object.armature_human_metarig_add()
            human_metarig = bpy.context.editable_objects[-1]

            ##################
            ### Scale armature

            bpy.ops.object.mode_set(mode="EDIT")

            for (
                    segment_name,
                    rigify_bones_list,
            ) in rigify_bone_to_skeleton_segment_name_correspondance.items():

                for rigify_bone_name in rigify_bones_list:
                    median_segment_length = skeleton_segment_lengths_dict[segment_name]["median"]
                    median_segment_length *= 0.001  # scale to meters

                    segment_length = median_segment_length / len(
                        rigify_bones_list
                    )  # divide by number of bones in segment (for now, eventually will want to set weights
                    print(f"setting {rigify_bone_name} length to: {segment_length:.3f} m")
                    human_metarig.data.edit_bones[rigify_bone_name].length = segment_length

            delete_bones = ["pelvis.L",
                            "pelvis.R",
                            "breast.L",
                            "breast.R",
                            "heel.02.L",
                            "heel.02.R",
                            "toe.L",
                            "toe.R", ]

            for bone in delete_bones:
                human_metarig.data.edit_bones.remove(human_metarig.data.edit_bones[bone])

            extra_scale_bones = {"spine.005": .2}

            for bone, scale in extra_scale_bones.items():
                human_metarig.data.edit_bones[bone].length *= scale

            ####
            #### Constrain bones to empties
            ####

            # loop through dictionary applying  constraints
            bpy.ops.object.mode_set(mode="POSE")

            for bone_name, bone_constraint_dict in rig_constraint_dict_of_dicts.items():
                print(f"---Setting constraints for bone:{bone_name}---")
                apply_constraints_to_bone(
                    bone_name=bone_name,
                    bone_constraint_dict=bone_constraint_dict,
                    armature_rig=human_metarig,
                )

            print(
                "____________________________________________________________________________\n",
                "Done constraining armature bones to follow keyframed empties!",
                "____________________________________________________________________________",
            )


        except Exception as e:
            print("Error creating rig!! (You might need to manually activate the `rigify` addon in Blender)")
            print(e)

    for window in bpy.context.window_manager.windows:
        for area in window.screen.areas:  # iterate through areas in current screen
            if area.type == "VIEW_3D":
                for space in area.spaces:  # iterate through spaces in current VIEW_3D area
                    if space.type == "VIEW_3D":  # check if space is a 3D view
                        space.shading.type = "MATERIAL"

    # save .blend file
    bpy.ops.object.mode_set(mode="OBJECT")
    bpy.ops.wm.save_as_mainfile(filepath=str(blender_file_save_path))
    print(f"Saved .blend file to: {blender_file_save_path}")

    print("try to hide the shameful metarig and then save again lol")
    human_metarig.hide_set = True
    bpy.ops.wm.save_as_mainfile(filepath=str(blender_file_save_path))


    print(
        "____________________________________________________________________________\n",
        "____________________________________________________________________________\n",
        "Done creating Blender scene!\n",
        f"Saved .blend file to: {blender_file_save_path}\n",
        "You can now open the `.blend` file in Blender and play the animation!\n",
        "____________________________________________________________________________\n",
        "____________________________________________________________________________",
    )


def add_videos_to_scene(videos_path: Union[Path, str], parent_object: bpy.types.Object, vid_location_scale: float = 2):
    print("loading videos as planes...")

    number_of_videos = len(list(get_video_paths(videos_path)))
    print(f"Found {number_of_videos} videos in {videos_path}")
    for (
            video_number,
            video_path,
    ) in enumerate(get_video_paths(videos_path)):
        print(f"Adding video: {video_path.name} to scene")

        bpy.ops.import_image.to_plane(
            files=[{"name": video_path.name}],
            directory=str(video_path.parent),
            shader="EMISSION",
        )
        print(f"Added video: {video_path.name} to scene")
        video_as_plane = bpy.context.editable_objects[-1]
        print(f"video_as_plane: {video_as_plane}")
        video_as_plane.name = "video_" + str(video_number)
        print(f"video_as_plane.name: {video_as_plane.name}")
        buffer = 1.1
        vid_x = (video_number * buffer - np.mean(np.arange(0, number_of_videos))) * vid_location_scale

        video_as_plane.location = [
            vid_x,
            vid_location_scale,
            1,
        ]
        video_as_plane.rotation_euler = [np.pi / 2, 0, 0]
        video_as_plane.scale = [1] * 3
        video_as_plane.parent = parent_object


def create_keyframed_empty_from_3d_trajectory_data(
        trajectory_fr_xyz: np.ndarray,
        trajectory_name: str,
        parent_origin: bpy.types.Object = None,
        empty_scale: float = 0.1,
        empty_type: str = "PLAIN_AXES",
):
    """
    Create a key framed empty from 3d trajectory data
    """
    print(f"Creating keyframed empty from: {trajectory_name}...")
    bpy.ops.object.empty_add(type=empty_type)
    empty_object = bpy.context.editable_objects[-1]
    empty_object.name = trajectory_name

    empty_object.scale = [empty_scale] * 3

    empty_object.parent = parent_origin

    for frame_number in range(trajectory_fr_xyz.shape[0]):
        empty_object.location = [
            trajectory_fr_xyz[frame_number, 0],
            trajectory_fr_xyz[frame_number, 1],
            trajectory_fr_xyz[frame_number, 2],
        ]

        empty_object.keyframe_insert(data_path="location", frame=frame_number)


def put_sphere_meshes_on_empties(
        empty_names_list: list,
        parent_object: bpy.types.Object,
        sphere_scale: float = 0.02,
):
    """
    put uv sphere meshes on the empties in `empty_names_list` with a scale of `sphere_scale`
    """

    for empty_name in empty_names_list:
        bpy.ops.mesh.primitive_uv_sphere_add(segments=8, ring_count=8, scale=(sphere_scale, sphere_scale, sphere_scale))
        sphere = bpy.context.editable_objects[-1]
        sphere.name = f"{empty_name}_sphere"
        sphere.parent = parent_object

        constraint = sphere.constraints.new(type="COPY_LOCATION")
        constraint.target = bpy.data.objects[empty_name]


def create_stick_meshes_between_emtpies(body_frame_marker_xyz: np.ndarray,
                                        empty_names: List[str],
                                        body_connections: Dict[int, int],
                                        parent_object: bpy.types.Object, ):
    """make cylinder meshes between the empties as defined in `body_connections`"""
    for origin_marker_index, target_marker_index in body_connections.items():
        origin_marker_name = empty_names[origin_marker_index]
        target_marker_name = empty_names[target_marker_index]

        stick = create_stick_mesh()
        stick = bpy.context.editable_objects[-1]
        stick.name = f"{origin_marker_name}_to_{target_marker_name}"
        stick.parent = parent_object

        constraint = stick.constraints.new(type="COPY_LOCATION")
        constraint.target = bpy.data.objects[origin_marker_name]

        constraint = stick.constraints.new(type="DAMPED_TRACK")
        constraint.target = bpy.data.objects[target_marker_name]
        constraint.track_axis = "TRACK_Z"

        for frame_number in range(body_frame_marker_xyz.shape[0]):
            distance_between_markers = np.linalg.norm(
                body_frame_marker_xyz[frame_number, origin_marker_index, :] -
                body_frame_marker_xyz[frame_number, target_marker_index, :]
            )
            stick.scale = [1, 1, distance_between_markers]
            stick.keyframe_insert(data_path="scale", frame=frame_number)


def create_stick_mesh(vertices: int = 12,
                      radius: float = 0.02,
                      depth: float = 1.0,
                      location=(0, 0, 0),
                      rotation=(0, 0, 0),
                      scale=(1, 1, 1),
                      end_fill_type='TRIFAN',
                      enter_editmode=False,
                      align='WORLD'):
    bpy.ops.mesh.primitive_cylinder_add(radius=radius,
                                        depth=depth,
                                        vertices=vertices,
                                        location=location,
                                        rotation=rotation,
                                        scale=scale,
                                        end_fill_type=end_fill_type,
                                        enter_editmode=enter_editmode,
                                        align=align)
    # give 3dcursor new coordinates
    bpy.context.scene.cursor.location = (0.0, 0.0, -0.5)

    # set the origin on the current object to the 3dcursor location
    bpy.ops.object.origin_set(type='ORIGIN_CURSOR')
    return bpy.context.editable_objects[-1]


def apply_constraints_to_bone(bone_name: str, bone_constraint_dict: dict, armature_rig):
    """
    Apply constraints to a bone (that exists in `armature_rig` based on a dictionary of constraints
    """
    for (
            constraint_name,
            constrain_parameters_dict,
    ) in bone_constraint_dict.items():
        constraint_name = constraint_name.split(".")[
            0
        ]  # for duplicated constraints, named `[constraint_name].001`, etc
        print(f"constraint: {constraint_name} with parameters:{constrain_parameters_dict}")

        bone = armature_rig.pose.bones[bone_name]
        print(f"bone: {bone.name}")

        constraint = bone.constraints.new(type=constraint_name)
        print(f"constraint: {constraint.name}")

        # generic constraint settings
        if "target" in constrain_parameters_dict:

            if constrain_parameters_dict["target"] == "armature":
                constraint.target = armature_rig
                constraint.subtarget = constrain_parameters_dict["subtarget"]
                print(f"constraint.target: {constraint.target.name}")
                print(f"constraint.subtarget: {constraint.target.name}")
            else:
                constraint.target = bpy.data.objects[
                    constrain_parameters_dict["target"]
                ]  # point constraint at relevant empty object
                print(f"constraint.target: {constraint.target.name}")

        if "influence" in constrain_parameters_dict:
            constraint.influence = constrain_parameters_dict["influence"]
            print(f"constraint.influence: {constraint.influence}")

        # constraint-specific settings
        if constraint_name == "IK":
            constraint.chain_count = constrain_parameters_dict["chain_length"]
            print(f"constraint.chain_count: {constraint.chain_count}")

        if constraint_name == "LOCKED_TRACK":
            constraint.track_axis = constrain_parameters_dict["track_axis"]
            constraint.lock_axis = constrain_parameters_dict["lock_axis"]
            print(f"constraint.track_axis: {constraint.track_axis}")
            print(f"constraint.lock_axis: {constraint.lock_axis}")

        if constraint_name == "DAMPED_TRACK":
            if "track_axis" in constrain_parameters_dict:
                constraint.track_axis = constrain_parameters_dict["track_axis"]
                print(f"constraint.track_axis: {constraint.track_axis}")


def test_virtual_marker_definitions(virtual_marker_definitions_dict: dict):
    """
    Validate the virtual marker definitions dictionary to ensure that there are the same number of marker names and weights, and that the weights sum to 1
    """

    for (
            virtual_marker_name,
            virtual_marker_definition,
    ) in virtual_marker_definitions_dict.items():
        assert len(virtual_marker_definition["marker_names"]) == len(
            virtual_marker_definition["marker_weights"]
        ), f"marker_names and marker_weights must be the same length for virtual marker {virtual_marker_name}"
        assert (
                sum(virtual_marker_definition["marker_weights"]) == 1
        ), f"marker_weights must sum to 1 for virtual marker {virtual_marker_name}"


def calculate_virtual_marker_trajectory(
        trajectory_3d_frame_marker_xyz: np.ndarray,
        all_trajectory_names: list,
        component_trajectory_names: List,
        trajectory_weights: List,
) -> np.ndarray:
    """
    Create a virtual marker from a set of component markers. A 'Virtual Marker' is a 'fake' marker created by combining the data from 'real' (measured) marker/trajectory data
    `trajectory_3d_frame_marker_xyz`: all trajectory data in a numpy array with shape [frame, marker, xyz]
    `all_trajectory_names`: list of all trajectory names
    `component_trajectory_names`: the trajectories we'll use to make this virtual marker
    `trajectory_weights`: the weights we'll use to combine the compoenent trajectories into the virtual maker
    """

    # double check that the weights scale to one, otherwise this function will return screwy results
    assert np.sum(trajectory_weights) == 1, "Error - Trajectory_weights must sum to 1!"
    assert len(component_trajectory_names) == len(trajectory_weights), \
        "Error - component_trajectory_names and trajectory_weights must be the same length!"
    virtual_marker_xyz = np.zeros((trajectory_3d_frame_marker_xyz.shape[0], 3))

    for trajectory_name, weight in zip(component_trajectory_names, trajectory_weights):
        # pull out the trajectory data for this component trajectory and scale by it's `weight`
        component_trajectory_xyz = copy(
            trajectory_3d_frame_marker_xyz[:, all_trajectory_names.index(trajectory_name), :] * weight)

        # add it to the virtual marker
        virtual_marker_xyz += component_trajectory_xyz

    return virtual_marker_xyz


def creating_stick_figure_mesh_from_bone_dictionary(trajectory_frame_marker_xyz: np.ndarray,
                                                    body_connections: Dict[int, int]):
    print("creating_stick_figure_mesh_from_bone_dictionary")

    assert (
                len(trajectory_frame_marker_xyz.shape) == 3), "trajectory_frame_marker_xyz must have three axes - [frame, marker, xyz] (e.g. skeleton data from a single frame"
    mesh = bpy.data.meshes.new("stick_figure_mesh")

    init_zero_vertices = np.zeros((trajectory_frame_marker_xyz.shape[1], 3))

    edges = []
    for head_index, tail_index in body_connections.items():
        edges.append((head_index, tail_index))

    mesh.from_pydata(init_zero_vertices, edges, [])
    mesh.validate(verbose=True)
    mesh.update()

    skin_radius = .025
    print(f"Skinning stick figure mesh with radius {skin_radius}...")
    bpy.data.objects['stick_figure_mesh'].modifiers.new('stick_figure_mesh_skin', 'SKIN')
    bpy.ops.transform.skin_resize(value=(skin_radius, skin_radius, skin_radius))

    for frame_number in range(trajectory_frame_marker_xyz.shape[0]):
        for vertex in mesh.vertices:
            vertex.co = trajectory_frame_marker_xyz[frame_number, vertex.index, :]
            vertex.keyframe_insert(data_path="co", frame=frame_number)  # keyframe the vertex location
        mesh.update()

def get_video_paths(path_to_video_folder: Path) -> list:
    """Search the folder for 'mp4' files (case insensitive) and return them as a list"""

    list_of_video_paths = list(Path(path_to_video_folder).glob("*.mp4")) + list(
        Path(path_to_video_folder).glob("*.MP4")
    )
    unique_list_of_video_paths = get_unique_list(list_of_video_paths)

    return unique_list_of_video_paths


def get_unique_list(list: list) -> list:
    """Return a list of the unique elements from input list"""
    unique_list = []
    [unique_list.append(clip) for clip in list if clip not in unique_list]

    return unique_list


mediapipe_empty_names = {
    # generated by: freemocap/core_processes/detecting_things_in_2d_images/mediapipe_stuff/mediapipe_skeleton_names_and_connections.py
    'body': ['nose',
             'left_eye_inner',
             'left_eye',
             'left_eye_outer',
             'right_eye_inner',
             'right_eye',
             'right_eye_outer',
             'left_ear',
             'right_ear',
             'mouth_left',
             'mouth_right',
             'left_shoulder',
             'right_shoulder',
             'left_elbow',
             'right_elbow',
             'left_wrist',
             'right_wrist',
             'left_pinky',
             'right_pinky',
             'left_index',
             'right_index',
             'left_thumb',
             'right_thumb',
             'left_hip',
             'right_hip',
             'left_knee',
             'right_knee',
             'left_ankle',
             'right_ankle',
             'left_heel',
             'right_heel',
             'left_foot_index',
             'right_foot_index'],
    'hand': ['wrist',
             'thumb_cmc',
             'thumb_mcp',
             'thumb_ip',
             'thumb_tip',
             'index_finger_mcp',
             'index_finger_pip',
             'index_finger_dip',
             'index_finger_tip',
             'middle_finger_mcp',
             'middle_finger_pip',
             'middle_finger_dip',
             'middle_finger_tip',
             'ring_finger_mcp',
             'ring_finger_pip',
             'ring_finger_dip',
             'ring_finger_tip',
             'pinky_mcp',
             'pinky_pip',
             'pinky_dip',
             'pinky_tip'],

    'face': ['right_eye',
             'left_eye',
             'nose_tip',
             'mouth_center',
             'right_ear_tragion',
             'left_ear_tragion']}

# NOTE - numbers refer to the index of the name in the relevant list above
body_connections = {15: 19,
                    16: 18,
                    18: 20,
                    3: 7,
                    14: 16,
                    23: 24,
                    28: 32,
                    11: 12,
                    27: 29,
                    6: 8,
                    24: 26,
                    4: 5,
                    5: 6,
                    29: 31,
                    12: 14,
                    0: 4,
                    9: 10,
                    1: 2,
                    30: 32,
                    25: 27,
                    26: 28,
                    17: 19,
                    2: 3,
                    13: 15}

hand_connections = {3: 4,
                    0: 1,
                    17: 18,
                    13: 17,
                    18: 19,
                    5: 9,
                    14: 15,
                    9: 13,
                    1: 2,
                    10: 11,
                    19: 20,
                    6: 7,
                    15: 16,
                    2: 3,
                    11: 12,
                    7: 8}
# boy howdy mediapipe did not make it easy to get this info...
face_contour_marker_indices = [0, 7, 10, 13, 14, 17, 21, 33, 37, 39, 40, 46, 52, 53, 54, 55, 58, 61, 63, 65, 66, 67, 70,
                               78, 80, 81, 82, 84, 87, 88, 91, 93, 95, 103, 105, 107, 109, 127, 132, 133, 136, 144, 145,
                               146, 148, 149, 150, 152, 153, 154, 155, 157, 158, 159, 160, 161, 162, 163, 172, 173, 176,
                               178, 181, 185, 191, 234, 246, 249, 251, 263, 267, 269, 270, 276, 282, 283, 284, 285, 288,
                               291, 293, 295, 296, 297, 300, 308, 310, 311, 312, 314, 317, 318, 321, 323, 324, 332, 334,
                               336, 338, 356, 361, 362, 365, 373, 374, 375, 377, 378, 379, 380, 381, 382, 384, 385, 386,
                               387, 388, 389, 390, 397, 398, 400, 402, 405, 409, 415, 454, 466]

mediapipe_virtual_marker_definitions_dict = {
    "head_center": {
        "marker_names": ["left_ear", "right_ear"],
        "marker_weights": [0.5, 0.5],
    },
    "neck_center": {
        "marker_names": ["left_shoulder", "right_shoulder"],
        "marker_weights": [0.5, 0.5],
    },
    "trunk_center": {
        "marker_names": ["left_shoulder", "right_shoulder", "left_hip", "right_hip"],
        "marker_weights": [0.25, 0.25, 0.25, 0.25],
    },
    "hips_center": {
        "marker_names": ["left_hip", "right_hip"],
        "marker_weights": [0.5, 0.5],
    },
}

# Skeleton segments names and the rigify bone names that they correspond to
rigify_bone_to_skeleton_segment_name_correspondance = {
    "lower_spine": ["spine", "spine.001"],
    "upper_spine": ["spine.002", "spine.003"],
    "head": ["spine.004", "spine.005", "spine.006"],
    "left_hand": ["hand.L"],
    "left_forearm": ["forearm.L"],
    "left_upper_arm": ["upper_arm.L"],
    "right_hand": ["hand.R"],
    "right_forearm": ["forearm.R"],
    "right_upper_arm": ["upper_arm.R"],
    "left_thigh": ["thigh.L"],
    "left_calf": ["shin.L"],
    "left_foot": ["foot.L"],
    "right_thigh": ["thigh.R"],
    "right_calf": ["shin.R"],
    "right_foot": ["foot.R"],
}

# A dictionary where the `keys` are  rigify bone names and the `values` are bone constraints and their parameters.
# NOTE - the `targets` must correspond to the `mediapipe_body_names` list above (or one of the virtural markers we'll make later)
rig_constraint_dict_of_dicts = {
    "spine": {
        "COPY_LOCATION": {"target": "hips_center"},
    },
    "spine.003": {
        "IK": {"target": "neck_center", "chain_length": 4},
    },
    "spine.004": {
        "COPY_LOCATION": {"target": "neck_center"},
    },
    "spine.006": {
        "IK": {"target": "head_center", "chain_length": 2},
    },
    "face": {
        "COPY_LOCATION": {"target": "armature", "subtarget": "spine.006"},
        "DAMPED_TRACK": {
            "target": "head_center",
        },
        "LOCKED_TRACK": {
            "target": "nose",
            "track_axis": "TRACK_Z",
            "lock_axis": "LOCK_Y",
            "influence": 1.0,
        },
        "LOCKED_TRACK.001": {
            "target": "left_ear",
            "track_axis": "TRACK_Z",
            "lock_axis": "LOCK_Z",
            "influence": 1.0,
        },
    },
    "shoulder.R": {
        "COPY_LOCATION": {"target": "neck_center"},
        "DAMPED_TRACK": {"target": "right_shoulder"},
    },
    "upper_arm.R": {
        "DAMPED_TRACK": {"target": "right_elbow"},
    },
    "forearm.R": {
        "DAMPED_TRACK": {"target": "right_wrist"},
    },
    "hand.R": {
        "DAMPED_TRACK": {"target": "right_index"},
    },
    "shoulder.L": {
        "COPY_LOCATION": {"target": "neck_center"},
        "DAMPED_TRACK": {"target": "left_shoulder"},
    },
    "upper_arm.L": {
        "DAMPED_TRACK": {"target": "left_elbow"},
    },
    "forearm.L": {
        "DAMPED_TRACK": {"target": "left_wrist"},
    },
    "hand.L": {
        "DAMPED_TRACK": {"target": "left_index"},
    },
    "thigh.R": {
        "COPY_LOCATION": {"target": "right_hip"},
        "DAMPED_TRACK": {"target": "right_knee"},
    },
    "shin.R": {
        "DAMPED_TRACK": {"target": "right_ankle"},
    },
    "foot.R": {
        "DAMPED_TRACK": {"target": "right_foot_index"},
    },
    "thigh.L": {
        "COPY_LOCATION": {"target": "left_hip"},
        "DAMPED_TRACK": {"target": "left_knee"},
    },
    "shin.L": {
        "DAMPED_TRACK": {"target": "left_ankle"},
    },
    "foot.L": {
        "DAMPED_TRACK": {"target": "left_foot_index"},
    },
}

print(f"__name__: {__name__}")

if __name__ == "__main__" or __name__ == "<run_path>":

    # this is the part that actually runs the script
    argv = sys.argv
    create_rig_input = False

    print(f"Running script to create Blender file from freemocap session data from script {__file__}")
    try:
        # Session path
        # get session path as command line argument
        print(f"Received command line arguments: {argv}")
        argv = argv[argv.index("--") + 1:]

        recording_path_input = Path(argv[0])
        blender_file_save_path_input = Path(argv[1])
        create_rig_input = bool(argv[2])
    except Exception as e:
        print(f"There was a problem loading command line arguments: {e}. Using hardcoded values instead.")

        if len(argv) == 0:

            ####
            # Change the path below (`recording_path_input`) to the path to your recording session folder
            ####
            recording_path_input = Path(
                r"C:\Users\jonma\freemocap_data\recording_sessions\session_2023-02-15_08_46_43_wud\recording_08_47_25_gmt-5_wud"
            )
            blender_file_save_path_input = Path(recording_path_input) / f"{Path(recording_path_input).name}_.blend"
            create_rig = True

    print(f"recording_path_input: {recording_path_input}, \n"
          f"blender_file_save_path_input: {blender_file_save_path_input}, \n"
          f"create_rig_input: {create_rig_input}")

    main(recording_path=recording_path_input,
         blender_file_save_path=blender_file_save_path_input,
         mediapipe_empty_names=mediapipe_empty_names,
         create_rig=create_rig_input,)



import logging
import platform
from pathlib import Path
from typing import Union


logger = logging.getLogger(__name__)


def guess_blender_exe_path_from_path(base_path: Union[str, Path]) -> Path:
    blender_folder_path = [path for path in base_path.glob("Blender*")]
    if len(blender_folder_path) > 0:
        blender_exe_path = blender_folder_path[-1] / "blender.exe"
        return blender_exe_path


def get_best_guess_of_blender_path():
    if platform.system() == "Windows":

        base_path = Path("C:\Program Files\Blender Foundation")
        blender_exe_path = guess_blender_exe_path_from_path(base_path)

        if blender_exe_path is not None:
            if not blender_exe_path.is_file():
                base_path = Path(Path.home()) / "Blender Foundation"
                blender_exe_path = guess_blender_exe_path_from_path(base_path)

        if blender_exe_path is not None and blender_exe_path.is_file():
            logger.info(f"Windows machine detected - guessing that `blender` is installed at: {str(blender_exe_path)}")

            return str(blender_exe_path)
        else:
            logger.warning(
                f"Could not find `blender.exe` in the expected locations. Please located it manually (or install Blender, if it isn't installed)."
            )
            return None

    else:
        logger.info(f"Non-Windows machine detected - TODO - Test how this works on Mac/Linux")
        return None



import os
import subprocess
import sys


def open_file(filename):
    if sys.platform == "win32":
        os.startfile(filename)
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, filename])



import logging
import signal
import sys
from pathlib import Path

from PyQt6.QtCore import QTimer
from PyQt6.QtWidgets import QApplication

from freemocap.gui.qt.main_window.freemocap_main_window import FreemocapMainWindow, EXIT_CODE_REBOOT
from freemocap.gui.qt.utilities.get_qt_app import get_qt_app
from freemocap.system.paths_and_files_names import (
    get_freemocap_data_folder_path,
)
from freemocap.system.user_data.pipedream_pings import PipedreamPings

repo = Path(__file__).parent.parent.parent.parent
sys.path.insert(0, str(repo))

logger = logging.getLogger(__name__)


def sigint_handler(*args):
    """Handler for the SIGINT signal."""
    QApplication.quit()


def qt_gui_main():
    logger.info("Starting main...")
    signal.signal(signal.SIGINT, sigint_handler)
    app = get_qt_app()
    timer = QTimer()
    timer.start(500)

    pipedream_pings = PipedreamPings()

    while True:
        freemocap_main_window = FreemocapMainWindow(freemocap_data_folder_path=get_freemocap_data_folder_path(),
                                                    pipedream_pings=pipedream_pings)
        freemocap_main_window.show()
        timer.timeout.connect(freemocap_main_window.update)
        error_code = app.exec()
        logger.info(f"`main` exited with error code: {error_code}")
        freemocap_main_window.close()

        if not error_code == EXIT_CODE_REBOOT:
            print(f"Thank you for using freemocap \U0001F480 \U00002764 \U00002728")
            break

        logger.info("`main` exited with the 'reboot' code, so let's reboot!")

    sys.exit()


if __name__ == "__main__":
    qt_gui_main()



import logging
import multiprocessing
import threading
from logging.handlers import QueueHandler
from queue import Queue

from PyQt6 import QtCore
from PyQt6.QtCore import QThread
from PyQt6.QtWidgets import QApplication, QPlainTextEdit

logger = logging.getLogger(__name__)

log_view_logging_format_string = (
    "[%(asctime)s.%(msecs)04d] [%(levelname)4s] [%(name)s:%(funcName)s():%(lineno)s]\n :::::: %(message)s\n."
)

log_view_logging_formatter = logging.Formatter(fmt=log_view_logging_format_string, datefmt="%Y-%m-%dT%H:%M:%S")


class LoggingQueueListener(QThread):
    log_message_signal = QtCore.pyqtSignal(str)

    def __init__(self, logging_queue: Queue, exit_event: threading.Event, parent=None):
        super().__init__(parent)
        self._logging_queue = logging_queue
        self._exit_event = exit_event

    def run(self):
        logger.info("Starting LoggingQueueListener thread")
        try:
            while not self._exit_event.is_set():
                if self._logging_queue.empty():
                    continue

                record = self._logging_queue.get(block=True)

                if record is None:
                    break
                self.log_message_signal.emit(record.message)
        except Exception as e:

            self.close()

    def close(self):
        logger.info("Closing LoggingQueueListener thread")
        self._exit_event.set()


class LogViewWidget(QPlainTextEdit):
    # adapted from - https://stackoverflow.com/a/63853259/14662833
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setReadOnly(True)

        self._logging_queue = multiprocessing.Queue(-1)
        self._queue_handler = QueueHandler(self._logging_queue)
        self._queue_handler.setFormatter(log_view_logging_formatter)
        logging.getLogger("").handlers.append(self._queue_handler)

        self._exit_event = threading.Event()
        self._logging_queue_listener = LoggingQueueListener(
            logging_queue=self._logging_queue, exit_event=self._exit_event
        )
        self._logging_queue_listener.log_message_signal.connect(self.add_log)

        self._logging_queue_listener.start()

    def add_log(self, message):
        self.appendPlainText(message.rstrip())

    def closeEvent(self, event):
        logger.info("Closing LogViewWidget")
        self._exit_event.set()
        self._logging_queue_listener.close()
        logging.getLogger("").handlers.remove(self._queue_handler)
        super().closeEvent(event)


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    log_view_widget = LogViewWidget()
    log_view_widget.show()
    sys.exit(app.exec())



import logging
from pathlib import Path
from typing import Union

from PyQt6.QtCore import pyqtSignal
from PyQt6.QtGui import QFileSystemModel
from PyQt6.QtWidgets import QVBoxLayout, QListWidget, QLabel, QFormLayout, QLineEdit, \
    QTreeView, QPushButton, QDialog

from freemocap.system.paths_and_files_names import get_recording_session_folder_path, SYNCHRONIZED_VIDEOS_FOLDER_NAME
from freemocap.system.start_file import open_file

no_files_found_string = "No '.mp4' video files found! \n \n Note - We only look for `.mp4` files (for now). If your videos are a different format, convert them to `mp4` via online tools like `www.cloudconvert.com`, or softwares like `HandBrake`, `ffmpeg` or any video editing software"


logger = logging.getLogger(__name__)

class ImportVideosWizard(QDialog):
    folder_to_save_videos_to_selected = pyqtSignal(list, str)
    def __init__(self,
                 import_videos_path: Union[str, Path],
                 parent=None):
        super().__init__(parent=parent)

        self.setWindowTitle("Import Videos")



        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._import_directory_view = self._create_import_directory_view(import_videos_path)
        self._layout.addWidget(self._import_directory_view)

        self._video_file_paths = [str(path) for path in Path(import_videos_path).glob('*.mp4')]

        self._video_file_list_view = self._create_video_file_list_widget()
        self._layout.addWidget(self._video_file_list_view)

        self._form_layout = QFormLayout()
        self._layout.addLayout(self._form_layout)

        self._folder_name = f"import_{Path(import_videos_path).name}"
        self._folder_name_line_edit = QLineEdit(parent=self)
        self._folder_name_line_edit.textChanged.connect(self._handle_folder_name_line_edit_changed)

        self._folder_name_line_edit.setPlaceholderText(self._folder_name)
        self._form_layout.addRow("Recording Name:", self._folder_name_line_edit)

        self._folder_where_videos_will_be_saved_to_label = QLabel(self._get_folder_videos_will_be_saved_to())
        self._form_layout.addRow("Videos will be saved to:", self._folder_where_videos_will_be_saved_to_label)

        self._continue_button = QPushButton("Continue")
        self._continue_button.isDefault()
        self._continue_button.clicked.connect(self._handle_continue_button_clicked)
        self._form_layout.addRow(self._continue_button)



    def _get_folder_videos_will_be_saved_to(self):
        return str(Path(get_recording_session_folder_path()) / self._folder_name / SYNCHRONIZED_VIDEOS_FOLDER_NAME)

    def _create_import_directory_view(self, import_videos_path: Union[str, Path]):

        self._file_system_model = QFileSystemModel()
        self._file_system_model.setRootPath(str(import_videos_path))

        self._tree_view_widget = QTreeView()
        self._layout.addWidget(self._tree_view_widget)
        self._tree_view_widget.setModel(self._file_system_model)
        self._tree_view_widget.setRootIndex(self._file_system_model.index(str(import_videos_path)))
        self._tree_view_widget.setColumnWidth(0, 250)
        self._tree_view_widget.setHeaderHidden(False)
        self._tree_view_widget.setAlternatingRowColors(True)
        self._tree_view_widget.setWindowTitle(str(import_videos_path))

        self._tree_view_widget.doubleClicked.connect(self._open_file)


    def _open_file(self):
        index = self._tree_view_widget.currentIndex()
        file_path = self._file_system_model.filePath(index)
        logger.info(f"Opening file from file_system_view_widget: {file_path}")
        open_file(file_path)
    def _create_video_file_list_widget(self):


        list_view = QListWidget()
        list_view.setWordWrap(True)
        if len(self._video_file_paths) == 0:
            list_view.addItem(no_files_found_string)
        else:
            list_view.addItems(self._video_file_paths)
        return list_view

    def _handle_folder_name_line_edit_changed(self, event):
        self._folder_name = self._folder_name_line_edit.text()
        self._folder_where_videos_will_be_saved_to_label.setText(self._get_folder_videos_will_be_saved_to())

    def _handle_continue_button_clicked(self, event):
        self.folder_to_save_videos_to_selected.emit(
            self._video_file_paths,
            self._get_folder_videos_will_be_saved_to())
        self.accept()

if __name__ == "__main__":
    # from PyQt6.QtWidgets import QApplication
    # import sys
    #
    # app = QApplication(sys.argv)
    import_videos_window = ImportVideosWizard(import_videos_path=Path().home())
    import_videos_window.exec()
    # import_videos_window.show()
    # sys.exit(app.exec())




import logging
import subprocess
from pathlib import Path
from typing import Union

# blender_exe_path = (
#     r"C:\Users\jonma\Blender Foundation\stable\blender-3.1.0-windows-x64\blender.exe"
# )
# blender_exe_path = r"C:\Users\jonma\Blender Foundation\Blender 3.1\blender.exe"
from freemocap.core_processes.detecting_things_in_2d_images.mediapipe_stuff.mediapipe_skeleton_names_and_connections import (
    mediapipe_names_and_connections_dict,
)
from freemocap.system.paths_and_files_names import OUTPUT_DATA_FOLDER_NAME
from freemocap.utilities.save_dictionary_to_json import save_dictionary_to_json

logger = logging.getLogger(__name__)


def call_blender_subprocess_cgtinker(
    recording_folder_path: Union[str, Path],
    blender_file_path: Union[str, Path],
    blender_exe_path: Union[str, Path],
):
    path_to_this_py_file = Path(__file__).parent.resolve()

    blender_script_path = path_to_this_py_file / "cgtinker_blendarmocap_load.py"

    # freemocap_blender_megascript_path = (
    #     path_to_this_py_file / "freemocap_blender_megascript_take2.py"
    # )

    try:
        blender_exe_path = Path(blender_exe_path)
        if not blender_exe_path.exists():
            raise FileNotFoundError(f"Could not find the blender executable at {blender_exe_path}")
    except Exception as e:
        logger.error(e)
        return

    command_list = [
        str(blender_exe_path),
        "--background",
        "--python",
        str(blender_script_path),
        "--",
        str(recording_folder_path),
        str(blender_file_path),
        "1",  # bind to rig
        "1",  # load synchronized_videos
        "9999",  # timeout
        "0",  # load raw data
        "1,",  # load quick
    ]

    logger.info(f"Starting `blender` sub-process with this command: \n {command_list}")

    blender_process = subprocess.Popen(command_list, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    while True:
        output = blender_process.stdout.readline()
        if blender_process.poll() is not None:
            break
        if output:
            print(output.strip().decode())

    if blender_process.returncode == 0:
        print("Blender returned an error:")
        print(blender_process.stderr.read().decode())
    logger.info(f"Done with blender stuff :D")


def call_blender_subprocess_alpha_megascript(
    recording_folder_path: Union[str, Path],
    blender_file_path: Union[str, Path],
    blender_exe_path: Union[str, Path],
    good_clean_frame_number: int = 0,
):
    path_to_this_py_file = Path(__file__).parent.resolve()

    freemocap_blender_megascript_path = (
        path_to_this_py_file / "blender_bpy_export_scripts" / "alpha_freemocap_blender_megascript.py"
    )

    # freemocap_blender_megascript_path = (
    #     path_to_this_py_file / "freemocap_blender_megascript_take2.py"
    # )

    try:
        blender_exe_path = Path(blender_exe_path)
        if not blender_exe_path.exists():
            raise FileNotFoundError(f"Could not find the blender executable at {blender_exe_path}")
    except Exception as e:
        logger.error(e)
        return

    command_list = [
        str(blender_exe_path),
        "--background",
        "--python",
        str(freemocap_blender_megascript_path),
        "--",
        str(recording_folder_path),
        str(blender_file_path),
        str(good_clean_frame_number),
    ]

    logger.info(f"Starting `blender` sub-process with this command: \n {command_list}")

    blender_process = subprocess.Popen(command_list, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    while True:
        output = blender_process.stdout.readline()
        if blender_process.poll() is not None:
            break
        if output:
            print(output.strip().decode())

    if blender_process.returncode == 0:
        print("Blender returned an error:")
        print(blender_process.stderr.read().decode())
    logger.info(f"Done with blender stuff :D")


def call_blender_subprocess_megascript_take2(
    recording_folder_path: Union[str, Path],
    blender_file_path: Union[str, Path],
    blender_exe_path: Union[str, Path],
):
    path_to_this_py_file = Path(__file__).parent.resolve()

    freemocap_blender_megascript_path = (
        path_to_this_py_file / "blender_bpy_export_scripts" / "freemocap_blender_megascript_take2.py"
    )

    if not (
        Path(recording_folder_path) / OUTPUT_DATA_FOLDER_NAME / "mediapipe_names_and_connections_dict.json"
    ).is_file():
        save_dictionary_to_json(
            save_path=str(Path(recording_folder_path) / OUTPUT_DATA_FOLDER_NAME),
            file_name="mediapipe_names_and_connections_dict.json",
            dictionary=mediapipe_names_and_connections_dict,
        )

    try:
        blender_exe_path = Path(blender_exe_path)
        if not blender_exe_path.exists():
            raise FileNotFoundError(f"Could not find the blender executable at {blender_exe_path}")
    except Exception as e:
        logger.error(e)
        return

    command_list = [
        str(blender_exe_path),
        "--background",
        "--python",
        str(freemocap_blender_megascript_path),
        "--",
        str(recording_folder_path),
        str(blender_file_path),
    ]

    logger.info(f"Starting `blender` sub-process with this command: \n {command_list}")

    blender_process = subprocess.Popen(command_list, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    while True:
        output = blender_process.stdout.readline()
        if blender_process.poll() is not None:
            break
        if output:
            print(output.strip().decode())

    if blender_process.returncode == 0:
        print("Blender returned an error:")
        print(blender_process.stderr.read().decode())
    logger.info(f"Done with blender stuff :D")



from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QToolBar, QWidget

from freemocap.gui.qt.widgets.control_panel.calibration_control_panel import CalibrationControlPanel
from freemocap.gui.qt.widgets.control_panel.process_mocap_data_panel.process_motion_capture_data_panel import (
    ProcessMotionCaptureDataPanel,
)


class ToolBar(QToolBar):
    def __init__(
        self,
        calibration_control_panel: CalibrationControlPanel,
        process_motion_capture_data_panel: ProcessMotionCaptureDataPanel,
        visualize_data_widget: QWidget,
        directory_view_widget: QWidget,
        parent=None,
    ):
        super().__init__(parent=parent)

        self._calibration_control_panel = calibration_control_panel
        self._process_motion_capture_data_panel = process_motion_capture_data_panel
        self._visualize_data_widget = visualize_data_widget
        self._directory_view_widget = directory_view_widget

        self.setMovable(True)
        self.setFloatable(True)
        self.setAllowedAreas(Qt.ToolBarArea.AllToolBarAreas)

        self._create_toolbar_actions()
        self._create_tool_bar()

    def _create_tool_bar(self):
        self.addAction(self._open_directory_action)

    def _create_toolbar_actions(self):
        self._open_directory_action = self.addAction("Open Directory")
        self._open_directory_action.triggered.connect(self._open_directory_widget)

    def _open_directory_widget(self):
        self._directory_view_widget.show()

        # self._directory_view_dock_widget.raise_()
        # self._directory_view_dock_widget.activateWindow()



from dataclasses import dataclass
from typing import Dict

import cv2


@dataclass
class CharucoBoardDefinition:
    try:
        # to make compatible across opencv 4.6 and 4.7
        aruco_marker_dict: Dict = cv2.aruco.Dictionary_get(cv2.aruco.DICT_4X4_250)
    except:
        aruco_marker_dict: Dict = cv2.aruco.getPredefinedDictionary(cv2.aruco.DICT_4X4_250)

    # number_of_squares_width: int = 5
    # number_of_squares_height: int = 3
    number_of_squares_width: int = 7
    number_of_squares_height: int = 5
    black_square_side_length: int = 1
    aruco_marker_length_proportional: float = 0.8

    def __post_init__(self):
        try:
            # opencv 4.6
            self.charuco_board = cv2.aruco.CharucoBoard_create(
                self.number_of_squares_width,
                self.number_of_squares_height,
                self.black_square_side_length,
                self.aruco_marker_length_proportional,
                self.aruco_marker_dict,
            )
        except:
            # opencv 4.7
            self.charuco_board = cv2.aruco.CharucoBoard(
                size=[self.number_of_squares_width, self.number_of_squares_height],
                squareLength=self.black_square_side_length,
                markerLength=self.aruco_marker_length_proportional,
                dictionary=self.aruco_marker_dict,
            )

        self.number_of_charuco_corners = (self.number_of_squares_width - 1) * (
                self.number_of_squares_height - 1
        )
